#!/usr/bin/env python3

"""

Copyright (c) 2016, Thorsten Bischler <thorsten.bischler@uni-wuerzburg.de>

Permission to use, copy, modify, and/or distribute this software for
any purpose with or without fee is hereby granted, provided that the
above copyright notice and this permission notice appear in all
copies.

THE SOFTWARE IS PROVIDED 'AS IS' AND THE AUTHOR DISCLAIMS ALL
WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE
AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.

"""

import sys
import argparse
import numpy as np
from scipy import stats
from wiggle import WiggleParser

__description__ = ("Sliding window peak calling script for identification of "
                   "protein binding sites based on RIP-seq data")
__author__ = "Thorsten Bischler <thorsten.bischler@uni-wuerzburg.de>"
__copyright__ = ("2016 by Thorsten Bischler "
                 "<thorsten.bischler@uni-wuerzburg.de>")
__license__ = "ISC license"
__email__ = "thorsten.bischler@uni-wuerzburg.de"
__version__ = "1.0"


def main():
    parser = argparse.ArgumentParser(description=__description__)
    parser.add_argument('-f', "--wt_file", required=True,
                        help="Path to control wiggle file for the respective "
                        "strand")
    parser.add_argument('-t', "--tag_file", required=True,
                        help="Path to coIP wiggle file for the respective "
                        "strand")
    parser.add_argument('-r', "--replicons_with_length", nargs='+',
                        help="List of replicon identifier (must match with "
                        "wiggle files) and length pairs in the form "
                        "'replicon_1:length_1 replicon_2:length_2 ...'")
    parser.add_argument('-w', "--window_size", type=int, default=25,
                        help="Applied window size")
    parser.add_argument('-l', "--step_size", type=int, default=5,
                        help="Applied step size")
    parser.add_argument('-c', "--fc_cutoff", type=float, default=5.0,
                        help="Minimum fold change for enrichment in coIP "
                        "library compared to control")
    parser.add_argument('-p', "--per_90_factor", type=float, default=3.0,
                        help="Factor to the 90th percentile of the coIP "
                        "library coverage used to define the minimum "
                        "expression cutoff")
    parser.add_argument('-s', "--strand", choices=('forward', 'reverse'),
                        required=True, help="Strand for which the analysis is "
                        "performed")
    args = parser.parse_args()
    fc = CalcFoldChange(args)
    replicon_dict = fc.calc_ratio_per_window()
    ae = AssembleEnrichment(args, replicon_dict)
    ae.def_enriched_regions()


class CalcFoldChange(object):

    def __init__(self, args):
        self._wiggle_parser = WiggleParser()
        self._wt_file = args.wt_file
        self._tag_file = args.tag_file
        self._replicons_with_length = args.replicons_with_length
        self._window_size = args.window_size
        self._step_size = args.step_size
        self._per_90_factor = args.per_90_factor
        self._replicon_dict = {}

    def _get_replicon_length(self):
        for replicon_with_length in self._replicons_with_length:
            replicon, length = replicon_with_length.split(':')
            print("Using replicon %s with %s bases" % (replicon, length))
            self._replicon_dict[replicon] = {'length': int(length)}

    def _store_expr(self):
        with open(self._wt_file) as wt_fh, open(self._tag_file) as tag_fh:
            for entry_wt, entry_tag in zip(
                    self._wiggle_parser.entries(wt_fh),
                    self._wiggle_parser.entries(tag_fh)):
                assert(entry_wt.chrom_name == entry_tag.chrom_name)
                if entry_wt.chrom_name not in self._replicon_dict:
                    sys.stderr.write("Missing length for replicon %s\n" %
                                     entry_wt.chrom_name)
                    sys.exit(1)
                pos_value_pairs_wt = dict(entry_wt.pos_value_pairs)
                pos_value_pairs_tag = dict(entry_tag.pos_value_pairs)
                if len(pos_value_pairs_wt) == 0 or len(
                        pos_value_pairs_tag) == 0:
                    print("%s: At least one replicon has no coverage for "
                          "this libs." % (entry_wt.chrom_name))
                    continue

                self._replicon_dict[entry_wt.chrom_name][
                    "wt_array"] = np.array([abs(pos_value_pairs_wt.get(
                        pos, 0.0)) for pos in range(1, self._replicon_dict[
                            entry_wt.chrom_name]['length'] + 1)])
                self._replicon_dict[entry_tag.chrom_name][
                    "tag_array"] = np.array([abs(pos_value_pairs_tag.get(
                        pos, 0.0)) for pos in range(1, self._replicon_dict[
                            entry_tag.chrom_name]['length'] + 1)])

                self._replicon_dict[entry_wt.chrom_name][
                    "per_90_wt"] = stats.scoreatpercentile(self._replicon_dict[
                        entry_wt.chrom_name]["wt_array"], 90)
                self._replicon_dict[entry_tag.chrom_name][
                    "per_90_tag"] = stats.scoreatpercentile(
                        self._replicon_dict[entry_tag.chrom_name][
                            "tag_array"], 90)

                print("%s: WT per_90 = %s, TAG per_90 = %s" % (
                    entry_wt.chrom_name,
                    self._replicon_dict[entry_wt.chrom_name]["per_90_wt"],
                    self._replicon_dict[entry_tag.chrom_name]["per_90_tag"]))

    def calc_ratio_per_window(self):
        self._get_replicon_length()
        self._store_expr()
        np.seterr(divide='ignore')
        for repl_values in self._replicon_dict.values():
            repl_values['window_dict'] = {}
            for start in range(1, repl_values['length'] - self._window_size +
                               self._step_size + 1, self._step_size):
                end = (start + self._window_size - 1)
                if end > repl_values['length']:
                    end = repl_values['length']

                value_wt = np.average(repl_values['wt_array'][start-1:end])
                value_tag = np.average(repl_values['tag_array'][start-1:end])

                if value_tag < (self._per_90_factor * repl_values[
                        'per_90_tag']):
                    repl_values['window_dict'][(start, end)] = 0.0
                else:
                    repl_values['window_dict'][(start, end)] = np.divide(
                        value_tag, value_wt)
        return self._replicon_dict


class AssembleEnrichment(object):

    def __init__(self, args, replicon_dict):
        self._fc_cutoff = args.fc_cutoff
        self._strand = args.strand
        self._replicon_dict = replicon_dict

    def def_enriched_regions(self):
        for replicon, repl_values in self._replicon_dict.items():
            repl_values['gff_list'] = []
            start, end = None, None
            er = False
            rightmost_er_pos = 0

            sorted_key_list = sorted(repl_values['window_dict'].keys())
            for index, key in enumerate(sorted_key_list):

                fold_change = repl_values['window_dict'][key]

                if fold_change >= self._fc_cutoff and not er:
                    if key[0] <= rightmost_er_pos:
                        pass
                    elif rightmost_er_pos > 0:
                        end = rightmost_er_pos
                        self._add_region(replicon, start, end)
                        start = key[0]
                    if start is None:
                        start = key[0]
                    rightmost_er_pos = key[1]

                elif fold_change >= self._fc_cutoff and er:
                    rightmost_er_pos = key[1]

                elif fold_change < self._fc_cutoff:
                    er = False

                if index == len(sorted_key_list) - 1 and start is not None:
                    end = rightmost_er_pos
                    self._add_region(replicon, start, end)
            self._write_gff_file(replicon)

    def _add_region(self, replicon, start, end):
        fold_change = np.divide(np.average(self._replicon_dict[replicon][
            'tag_array'][start-1:end]), np.average(self._replicon_dict[
                replicon]['wt_array'][start-1:end]))
        self._replicon_dict[replicon]['gff_list'].append(
            "%s\tsliding_window_peak_calling_script\tpeak_region\t%s\t%s\t.\t"
            "%s\t.\tID=er%s%s;fc=%s" % (
                replicon,
                start,
                end,
                '+' if self._strand == 'forward' else '-',
                len(self._replicon_dict[replicon]['gff_list']),
                '+' if self._strand == 'forward' else '-',
                fold_change)
        )

    def _write_gff_file(self, replicon):
        with open("%s_%s.gff" % (replicon, self._strand), 'w') as out_gff_fh:
            out_gff_fh.write(
                "##gff-version 3\n"
                "#!gff-spec-version 1.20\n"
                "##sequence-region %s 1 %s\n"
                "%s\n"
                "###\n" % (replicon,
                           self._replicon_dict[replicon]['length'],
                           '\n'.join(line for line in self._replicon_dict[
                               replicon]['gff_list'])))


if __name__ == "__main__":
    sys.exit(main())
